@@ -1,11 +1,12 @@
-#!/usr/bin/env python
 """Object mapper for Amazon DynamoDB.
 
 Based in part on mongokit's Document interface.
 
 Released under the GNU LGPL, version 3 or later (see COPYING).
 """
+from __future__ import absolute_import
 
+import datetime
 import logging
 import threading
 
@@ -74,37 +75,127 @@ class autoincrement_int(int):
     pass
 
 
-class json_list(list):
-    """JSON-serialized list data type for use in your schemas.
+_JSON_TYPES = frozenset([list, dict])
 
-    From your object's point of view, it's a list with arbitrary nesting and
-    whatever you want to store in it.
 
-    In DynamoDB itself, though, it's stored as a JSON-serialized string.
+class UTC(datetime.tzinfo):
+    """UTC timezone"""
+    def utcoffset(self, dt):
+        return datetime.timedelta(0)
+
+    def tzname(self, dt):
+        return "UTC"
+
+    def dst(self, dt):
+        return datetime.timedelta(0)
+
+
+utc_tz = UTC()
+
+
+def _get_proto_value(schema_type):
+    """Return a prototype value matching what schema_type will be serialized
+    as in DynamoDB:
+
+      - For strings and numbers, an instance of schema_type.
+      - For "special" types implemented at the mapper level (list, dict,
+        datetime), an empty string (this is what they're stored as in the DB).
     """
-    def __init__(self, source_str=""):
-        if source_str:
-            super(json_list, self).__init__(json.loads(source_str))
-        else:
-            super(json_list, self).__init__()
+    # Those types must be serialized as strings
+    if schema_type in _JSON_TYPES:
+        return u""
+
+    if schema_type == datetime.datetime:
+        return u""
+
+    # Regular string/number
+    return schema_type()
 
 
-class json_dict(dict):
-    """JSON-serialized dict data type for use in your schemas.
+def _get_default_value(schema_type):
+    """Return a default value matching schema_type:
 
-    From your object's point of view, it's a dict with arbitrary nesting and
-    whatever you want to store in it.
-
-    In DynamoDB itself, though, it's stored as a JSON-serialized string.
+      - For datetime.datetime, it's 1970-01-01 00:00:00 UTC.
+      - For container types, it's an empty container.
+      - For strings, it's an empty string.
+      - For numbers, it's zero.
     """
-    def __init__(self, source_str=""):
-        if source_str:
-            super(json_dict, self).__init__(json.loads(source_str))
-        else:
-            super(json_dict, self).__init__()
+    if schema_type == datetime.datetime:
+        # Beginning of the UNIX epoch
+        return datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=utc_tz)
+
+    return schema_type()
 
 
-_JSON_TYPES = frozenset([json_list, json_dict])
+def _python_to_dynamodb(value):
+    """Convert a Python object to a representation suitable to direct storage
+    in DynamoDB, according to a type from a DynamoDBModel schema.
+
+    If value should be represented as a missing value in DynamoDB
+    (empty string or set), None is returned.
+
+    ``_dynamodb_to_python(t, _python_to_dynamodb(v)) == v`` for any v.
+
+    :param value: The Python object to convert.
+
+    :return: ``value``, serialized to DynamoDB, or ``None`` if ``value`` must
+        be represented as a missing attribute.
+    """
+    if isinstance(value, tuple(_JSON_TYPES)):
+        # json serialization hooks for json_* data types.
+        return json.dumps(value, sort_keys=True)
+
+    if isinstance(value, datetime.datetime):
+        # datetime instances are stored as UTC in the DB itself.
+        # (that way, they become sortable)
+        # datetime objects without tzinfo are not supported.
+        return value.astimezone(utc_tz).strftime("%Y-%m-%dT%H:%M:%S.%f%z")
+
+    if value or value == 0:
+        return value
+
+    # Yes, that part is horrible. DynamoDB can't store empty
+    # sets/strings, so we're representing them as missing
+    # attributes on the DB side.
+    return None
+
+
+def _dynamodb_to_python(schema_type, value):
+    """Convert a DynamoDB attribute value to a Python object, according to a
+    type from a DynamoDBModel schema.
+
+    If value is None (usually because the attribute was missing in the first
+    place), a default value is returned (empty string or set, or 0, depending
+    on what the type is).
+
+    ``_dynamodb_to_python(t, _python_to_dynamodb(v)) == v`` for any v.
+
+    :param schema_type: A type supported by the mapper (TODO Clearly list those).
+
+    :param value: The DynamoDB attribute to convert to a Python object.
+        May be ``None``.
+
+    :return: An instance of ``schema_type``.
+    """
+    # This means a missing attribute for a _JSON_TYPES object is valid,
+    # and results in an empty sequence. Is that a bad thing?
+    if value is None:
+        # Empty sets/strings are represented as missing attributes
+        return schema_type()
+
+    if schema_type in _JSON_TYPES:
+        return schema_type(json.loads(value))
+
+    if schema_type == datetime.datetime:
+        # Parse TZ-aware isoformat
+
+        # strptime doesn't support timezone parsing (%z flag), so we're forcing
+        # the strings in the database to be UTC (+0000) for now.
+        # TODO Handle arbitrary timezones (with manual parsing).
+        return datetime.datetime.strptime(
+            value, "%Y-%m-%dT%H:%M:%S.%f+0000").replace(tzinfo=utc_tz)
+
+    return schema_type(value)
 
 
 class ConnectionBorg(object):
@@ -202,11 +293,12 @@ class ConnectionBorg(object):
 
         conn = self._get_connection()
         # It's a prototype/an instance, not a type.
-        hash_key_proto_value = hash_key_type()
+        hash_key_proto_value = _get_proto_value(hash_key_type)
         # None in the case of a hash-only table.
         if range_key_name:
             # We have a range key, its type must be specified.
-            range_key_proto_value = cls.__schema__[range_key_name]()
+            range_key_proto_value = _get_proto_value(
+                cls.__schema__[range_key_name])
         else:
             range_key_proto_value = None
 
@@ -267,8 +359,6 @@ class DynamoDBModel(object):
     TODO Add checks for common error cases:
         - Wrong datatypes in the schema
         - hash_key/range_key incorrectly defined
-
-    TODO Add consistent/eventually consistent reads flag.
     """
 
     # TODO Add checks to the various methods so that meaningful error messages
@@ -288,26 +378,41 @@ class DynamoDBModel(object):
         """
         instance = cls()
         for (name, type_) in cls.__schema__.iteritems():
-            try:
-                value = type_(d[name])
-            except KeyError:
-                # Fill in missing attributes with default values.
-                value = type_()
+            # Fill in missing attributes with default values.
+            value = _dynamodb_to_python(type_, d.get(name))
             setattr(instance, name, value)
         return instance
 
     @classmethod
-    def get(cls, hash_key_value, range_key_value=None):
+    def get(cls, hash_key_value, range_key_value=None, consistent_read=False):
         """Retrieve a single object from DynamoDB according to its primary key.
 
         Note that this is not a query method -- it will only return the object
         matching the exact primary key provided. Meaning that if the table is
         using a composite primary key, you need to specify both the hash and
         range key values.
+
+        :param hash_key_value: The value of the requested item's hash_key.
+
+        :param range_key_value: The value of the requested item's range_key,
+            if the table has a composite key.
+
+        :param consistent_read: If False (default), an eventually consistent
+            read is performed. Set to True for strongly consistent reads.
         """
         table = ConnectionBorg().get_table(cls.__table__)
+        # Convert the keys to DynamoDB values.
+        h_value = _python_to_dynamodb(hash_key_value)
+        if cls.__range_key__:
+            r_value = _python_to_dynamodb(range_key_value)
+        else:
+            r_value = None
+
         return cls.from_dict(
-            table.get_item(hash_key=hash_key_value, range_key=range_key_value))
+            table.get_item(
+                hash_key=h_value,
+                range_key=r_value,
+                consistent_read=consistent_read))
 
     @classmethod
     def get_batch(cls, keys):
@@ -321,11 +426,24 @@ class DynamoDBModel(object):
           - If the primary keys are composite (hash + range), keys must
             be a list of (hash_key, range_key) values
             (e.g. [("user1", 1), ("user1", 2), ("user1", 3)]).
+
+        get_batch *always* performs eventually consistent reads.
         """
         borg = ConnectionBorg()
         table = borg.get_table(cls.__table__)
+        # Convert all the keys to DynamoDB values.
+        if cls.__range_key__:
+            dynamo_keys = [
+                (
+                    _python_to_dynamodb(h),
+                    _python_to_dynamodb(r)
+                ) for (h, r) in keys
+            ]
+        else:
+            dynamo_keys = map(_python_to_dynamodb, keys)
+
         batch_list = borg.new_batch_list()
-        batch_list.add_batch(table, keys)
+        batch_list.add_batch(table, dynamo_keys)
 
         res = batch_list.submit()
         return [
@@ -333,7 +451,7 @@ class DynamoDBModel(object):
         ]
 
     @classmethod
-    def query(cls, hash_key_value, range_key_condition=None):
+    def query(cls, hash_key_value, range_key_condition=None, consistent_read=False):
         """Query DynamoDB for items matching the requested key criteria.
 
         You need to supply an exact hash key value, and optionally, conditions
@@ -350,12 +468,21 @@ class DynamoDBModel(object):
         :param range_key_condition: A condition instance from
             boto.dynamodb.condition -- one of EQ(x), LE(x), LT(x), GE(x),
             GT(x), BEGINS_WITH(x), BETWEEN(x, y).
+
+        :param consistent_read: If False (default), an eventually consistent
+            read is performed. Set to True for strongly consistent reads.
+
+        :rtype: generator
         """
         table = ConnectionBorg().get_table(cls.__table__)
-        return [
+        h_value = _python_to_dynamodb(hash_key_value)
+        return (
             cls.from_dict(d)
-            for d in table.query(hash_key_value, range_key_condition)
-        ]
+            for d in table.query(
+                h_value,
+                range_key_condition,
+                consistent_read=consistent_read)
+        )
 
     @classmethod
     def scan(cls, scan_filter=None):
@@ -369,12 +496,14 @@ class DynamoDBModel(object):
 
         :param scan_filter: A {attribute_name: condition} dict, where
             condition is a condition instance from boto.dynamodb.condition.
+
+        :rtype: generator
         """
         table = ConnectionBorg().get_table(cls.__table__)
-        return [
+        return (
             cls.from_dict(d)
             for d in table.scan(scan_filter)
-        ]
+        )
 
     def __init__(self):
         """Create a default empty instance of the class with default values
@@ -383,22 +512,18 @@ class DynamoDBModel(object):
         We're supplying this method to avoid the need for extra checks in save.
         """
         for (name, type_) in type(self).__schema__.iteritems():
-            setattr(self, name, type_())
+            setattr(self, name, _get_default_value(type_))
 
     def to_db_dict(self):
         """Return a dict representation of the object according to the class's
         schema, suitable for direct storage in DynamoDB.
         """
         out = {}
-        for (name, type_) in type(self).__schema__.iteritems():
-            value = getattr(self, name)
-            if type_ in _JSON_TYPES:
-                # json serialization hooks for json_* data types.
-                out[name] = json.dumps(value, sort_keys=True)
-            elif value or value == 0:
-                # Yes, that part is horrible. DynamoDB can't store empty
-                # sets/strings, so we're representing them as missing
-                # attributes on the DB side.
+        for name in type(self).__schema__:
+            value = _python_to_dynamodb(getattr(self, name))
+            if value is not None:
+                # DynamoDB can't store empty sets/strings, so we must
+                # represent them as missing attributes
                 out[name] = value
         return out
 
@@ -418,6 +543,10 @@ class DynamoDBModel(object):
             value = getattr(self, name)
             if isinstance(value, (set, frozenset)):
                 out[name] = list(value)
+            elif isinstance(value, datetime.datetime):
+                # Using strftime instead of the default str() representation
+                # because it outputs the zone as +HH:MM instead of +HHMM
+                out[name] = value.astimezone(utc_tz).strftime("%Y-%m-%dT%H:%M:%S.%f%z")
             else:
                 out[name] = value
         return out
@@ -490,15 +619,12 @@ class DynamoDBModel(object):
                 expected_db_values = {}
                 schema = type(self).__schema__
                 for (name, value) in expected_values.iteritems():
-                    type_ = schema[name]
-                    # This bit can probably be factored out with to_db_dict
-                    if type_ in _JSON_TYPES:
-                        expected_db_values[name] = json.dumps(value, sort_keys=True)
-                    elif value or value == 0:
-                        expected_db_values[name] = value
-                    else:
+                    db_value = _python_to_dynamodb(value)
+                    if db_value is None:
                         # Empty strings/sets must be represented as missing values
                         expected_db_values[name] = False
+                    else:
+                        expected_db_values[name] = db_value
             else:
                 # Forbid overwrites: do a conditional write on
                 # "this hash_key doesn't exist"
